/// @file   DpmAdapter.cpp
/// @brief  DPM wrapper.
/// @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
#include <serrno.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <dmlite/common/errno.h>
#include <dmlite/cpp/utils/urls.h>
#include <dpm_api.h>
#include <map>
#include <vector>

#include "Adapter.h"
#include "DpmAdapter.h"

using namespace dmlite;



DpmAdapterCatalog::DpmAdapterCatalog(unsigned retryLimit) throw (DmException):
  NsAdapterCatalog(retryLimit)
{
  dpm_client_resetAuthorizationId();
}



DpmAdapterCatalog::~DpmAdapterCatalog()
{
  // Nothing
}



std::string DpmAdapterCatalog::getImplId() const throw ()
{
  return std::string("DpmAdapterCatalog");
}



void DpmAdapterCatalog::setSecurityContext(const SecurityContext* ctx) throw (DmException)
{
  NsAdapterCatalog::setSecurityContext(ctx);
  
  // Call DPM API
  wrapCall(dpm_client_setAuthorizationId(boost::any_cast<uid_t>(ctx->user["uid"]),
                                         boost::any_cast<gid_t>(ctx->groups[0]["gid"]),
                                         "GSI",
                                         (char*)ctx->user.name.c_str()));

  if (this->nFqans_ > 0)
    wrapCall(dpm_client_setVOMS_data((char*)ctx->groups[0].name.c_str(),
                                     (char**)this->fqans_,
                                      this->nFqans_));
}



void DpmAdapterCatalog::unlink(const std::string& path) throw (DmException)
{
  int                    nReplies;
  struct dpm_filestatus *statuses;
  
  std::string            absolute;

  if (path[0] == '/')
    absolute = path;
  else
    absolute = this->cwdPath_ + "/" + path;

  struct stat stat = NsAdapterCatalog::extendedStat(absolute, false).stat;
  if (S_ISLNK(stat.st_mode)) {
    NsAdapterCatalog::unlink(absolute);
  }
  else {
    const char *path_c = absolute.c_str();
    RETRY(dpm_rm(1, (char**)&path_c, &nReplies, &statuses), this->retryLimit_);
    dpm_free_filest(nReplies, statuses);
  }
}



DpmAdapterPoolManager::DpmAdapterPoolManager(unsigned retryLimit,
                                             const std::string& passwd,
                                             bool useIp,
                                             unsigned life) throw (DmException):
  retryLimit_(retryLimit), tokenPasswd_(passwd), tokenUseIp_(useIp),
  tokenLife_(life), userId_(""), fqans_(NULL), nFqans_(0)
{
  // Nothing
}



DpmAdapterPoolManager::~DpmAdapterPoolManager()
{
  if (this->fqans_ != NULL) {
    for (size_t i = 0; i < this->nFqans_; ++i)
      delete [] this->fqans_[i];
    delete [] this->fqans_;
  }
}



std::string DpmAdapterPoolManager::getImplId() const throw()
{
  return std::string("DpmAdapter");
}



void DpmAdapterPoolManager::setStackInstance(StackInstance* si) throw (DmException)
{
  // Nothing
  this->si_ = si;
}



void DpmAdapterPoolManager::setSecurityContext(const SecurityContext* ctx) throw (DmException)
{
  // String => const char*
  if (this->fqans_ != NULL) {
    for (size_t i = 0; i < this->nFqans_; ++i)
      delete [] this->fqans_[i];
    delete [] this->fqans_;
  }
  
  this->nFqans_ = ctx->groups.size();
  this->fqans_  = new char* [this->nFqans_];
  for (size_t i = 0; i < this->nFqans_; ++i) {
    this->fqans_[i] = new char [ctx->groups[i].name.length() + 1];
    strcpy(this->fqans_[i], ctx->groups[i].name.c_str());
  }

  // Call DPM API
  wrapCall(dpm_client_setAuthorizationId(boost::any_cast<uid_t>(ctx->user["uid"]),
                                         boost::any_cast<gid_t>(ctx->groups[0]["gid"]),
                                         "GSI",
                                         (char*)ctx->user.name.c_str()));

  if (!ctx->groups.empty())
    wrapCall(dpm_client_setVOMS_data((char*)ctx->groups[0].name.c_str(),
                                     (char**)this->fqans_,
                                     ctx->groups.size()));
  
  if (this->tokenUseIp_)
    this->userId_ = ctx->credentials.remoteAddress;
  else
    this->userId_ = ctx->credentials.clientName;
}



std::vector<Pool> DpmAdapterPoolManager::getPools(PoolAvailability availability) throw (DmException)
{
  struct dpm_pool* dpmPools = NULL;
  int              nPools;

  try {
    // Get all pools
    RETRY(dpm_getpools(&nPools, &dpmPools), this->retryLimit_);
    
    std::vector<Pool> pools;
    Pool              pool;

    for (int i = 0; i < nPools; ++i) {
      pool.name = dpmPools[i].poolname;
      pool.type = "filesystem";
      pools.push_back(pool);
    }
    
    free(dpmPools);
    
    if (availability == kAny)
      return pools;
    
    // A pool is available if it has at least one fs available
    struct dpm_fs    *dpm_fs;
    std::vector<Pool> filtered;
  
    for (unsigned i = 0; i < pools.size(); ++i) {
      int  nFs;
      bool anyFsAvailable;

      if (dpm_getpoolfs((char*)pools[i].name.c_str(), &nFs, &dpm_fs) < 0)
        ThrowExceptionFromSerrno(serrno);

      anyFsAvailable = false;
      for (int j = 0; j < nFs && !anyFsAvailable; ++j) {
        
        switch (availability) {
          case kForWrite: case kForBoth:
            if (dpm_fs[j].status == 0)
              anyFsAvailable = true;
            break;
          default:
            if (dpm_fs[j].status != FS_DISABLED)
              anyFsAvailable = true;
        }

      }
      
      // Unless kNone, anyFsAvailable means a match
      if ((availability == kNone && !anyFsAvailable) ||
          (availability != kNone && anyFsAvailable))
        filtered.push_back(pools[i]);
      
      if (nFs > 0)
        free(dpm_fs);
    }
    
    return filtered;    
  }
  catch (...) {
    if (dpmPools != NULL)
      free(dpmPools);
    throw;
  }
}



Pool DpmAdapterPoolManager::getPool(const std::string& poolname) throw (DmException)
{
  std::vector<Pool> pools = this->getPools();
  
  for (unsigned i = 0; i < pools.size(); ++i) {
    if (poolname == pools[i].name)
      return pools[i];
  }
  
  throw DmException(DM_NO_SUCH_POOL, "Pool " + poolname + " not found");
}



void DpmAdapterPoolManager::newPool(const Pool&) throw (DmException)
{
  throw DmException(DM_NOT_IMPLEMENTED,
                    "DpmAdapterPoolManager::newPool not implemented");
}



void DpmAdapterPoolManager::updatePool(const Pool&) throw (DmException)
{
  throw DmException(DM_NOT_IMPLEMENTED,
                    "DpmAdapterPoolManager::updatePool not implemented");
}



void DpmAdapterPoolManager::deletePool(const Pool&) throw (DmException)
{
  throw DmException(DM_NOT_IMPLEMENTED,
                    "DpmAdapterPoolManager::deletePool not implemented");
}



Location DpmAdapterPoolManager::whereToRead(const std::string& path) throw (DmException)
{
  struct dpm_getfilereq     request;
  struct dpm_getfilestatus *statuses = NULL;
  int                       nReplies, wait;
  char                      r_token[CA_MAXDPMTOKENLEN + 1];
  std::string               rfn;
  char*                     rfio[] = {"rfio"};

  request.from_surl  = (char*)path.c_str();
  request.lifetime   = 0;
  request.f_type     = '\0';
  request.s_token[0] = '\0';
  request.ret_policy = '\0';
  request.flags      = 0;

  try {
    // Request
    RETRY(dpm_get(1, &request, 1, rfio, (char *)"libdm::dummy::dpm::get", 0,
                 r_token, &nReplies, &statuses), this->retryLimit_);
    if (nReplies < 1)
      throw DmException(DM_NO_REPLICAS, "No replicas found for " + path);

    // Wait for answer
    wait = statuses[0].status == DPM_QUEUED || statuses[0].status == DPM_RUNNING || statuses[0].status == DPM_ACTIVE;

    while (wait) {
      RETRY(dpm_getstatus_getreq(r_token, 1, &request.from_surl, &nReplies, &statuses), this->retryLimit_);
      if (!nReplies)
        throw DmException(DM_NO_REPLICAS, "No replicas found for " + path);

      switch(statuses[0].status) {
        case DPM_FAILED: case DPM_ABORTED: case DPM_EXPIRED:
          throw DmException(DM_QUERY_FAILED, "The DPM get request failed");
          break;
        case DPM_READY: case DPM_SUCCESS: case DPM_DONE:
          wait = 0;
          break;
      }
    }

    rfn = statuses[0].turl;
    dpm_free_gfilest(nReplies, statuses);
    
    // Build the answer
    Url rloc(rfn);
    rloc.path = Url::normalizePath(rloc.path);
    
    Chunk single;
    
    single.host = rloc.domain;
    single.path = rloc.path;
    
    single.offset = 0;
    single.size   = this->si_->getCatalog()->extendedStat(path).stat.st_size;
    
    single["token"] = dmlite::generateToken(this->userId_, rloc.path,
                                            this->tokenPasswd_, this->tokenLife_);
    
    return Location(1, single);          
  }
  catch (...) {
    // On exceptions, free first!
    if (statuses != NULL)
      dpm_free_gfilest(nReplies, statuses);
    throw;
  }
}



Location DpmAdapterPoolManager::whereToRead(ino_t) throw (DmException)
{
  throw DmException(DM_NOT_IMPLEMENTED, "DpmAdapter: Access by inode not supported");
}



Location DpmAdapterPoolManager::whereToWrite(const std::string& path) throw (DmException)
{
  struct dpm_putfilereq     reqfile;
  struct dpm_putfilestatus *statuses = NULL;
  int                       nReplies, wait;
  char                      token[CA_MAXDPMTOKENLEN + 1];
  std::string               spaceToken;
  char*                     rfio[] = {"rfio"};

  reqfile.to_surl        = (char*)path.c_str();
  reqfile.f_type         = 'P';
  reqfile.lifetime       = 0;
  reqfile.requested_size = 0;
  reqfile.f_lifetime     = 0;
  reqfile.ret_policy     = '\0';
  reqfile.ac_latency     = '\0';
  reqfile.s_token[0]     = '\0';
  
  try {
    spaceToken = boost::any_cast<std::string>(this->si_->get("SpaceToken"));
  }
  catch (...) {
  }

  if (!spaceToken.empty()) {
    char **space_ids;

    RETRY(dpm_getspacetoken(spaceToken.c_str(), &nReplies, &space_ids),
          this->retryLimit_);
    
    strncpy(reqfile.s_token, space_ids[0], sizeof(reqfile.s_token));

    for(int i = 0; i < nReplies; ++i)
      free(space_ids[i]);
    free(space_ids);
  }

  try {
    RETRY(dpm_put(1, &reqfile, 1, rfio, (char *)"libdm::dummy::dpm::put", 0,
                  0, token, &nReplies, &statuses), this->retryLimit_);

    wait = statuses[0].status == DPM_QUEUED  ||
           statuses[0].status == DPM_RUNNING ||
           statuses[0].status == DPM_ACTIVE;

    while (wait) {
      RETRY(dpm_getstatus_putreq(token, 1, &reqfile.to_surl, &nReplies, &statuses),
            this->retryLimit_);
      if (!nReplies)
        throw DmException(DM_PUT_ERROR, "Didn't get a destination from DPM");

      switch(statuses[0].status) {
        case DPM_FAILED: case DPM_ABORTED: case DPM_EXPIRED:
          throw DmException(DM_QUERY_FAILED, "The DPM put request failed");
          break;
        case DPM_READY: case DPM_SUCCESS:
          wait = 0;
          break;
      }
    }
    
    Url rloc(statuses[0].turl);
    dpm_free_pfilest(nReplies, statuses);
    
    rloc.path = Url::normalizePath(rloc.path);
    
    Chunk single;
    
    single.host = rloc.domain;
    single.path = rloc.path;
    
    single.offset = 0;
    single.size   = 0;
    
    single["sfn"]      = path;
    single["dpmtoken"] = std::string(token);
    single["token"]    = dmlite::generateToken(this->userId_, rloc.path,
                                               this->tokenPasswd_, this->tokenLife_, true);
    
    return Location(1, single);
  }
  catch (...) {
    // On exceptions, free first!
    if (statuses != NULL)
      dpm_free_pfilest(nReplies, statuses);
    throw;
  }
}
