/// @file   FilesystemDriver.cpp
/// @brief  Regular Filesystem pool
/// @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
#include <dmlite/cpp/dmlite.h>
#include <dmlite/cpp/utils/urls.h>
#include <dpm_api.h>
#include <errno.h>
#include <serrno.h>
#include <stdlib.h>
#include <string.h>

#include "Adapter.h"
#include "DpmAdapter.h"
#include "FilesystemDriver.h"


using namespace dmlite;



FilesystemPoolDriver::FilesystemPoolDriver(const std::string& passwd, bool useIp,
                                           unsigned life, unsigned retryLimit):
    secCtx_(NULL), tokenPasswd_(passwd), tokenUseIp_(useIp), tokenLife_(life),
    retryLimit_(retryLimit), fqans_(NULL), nFqans_(0)
{
  dpm_client_resetAuthorizationId();
}



FilesystemPoolDriver::~FilesystemPoolDriver()
{
  if (this->fqans_ != NULL) {
    for (int i = 0; i < this->nFqans_; ++i)
      delete [] this->fqans_[i];
    delete [] this->fqans_;
  }
}



std::string FilesystemPoolDriver::getImplId() const throw ()
{
  return "FilesystemPoolDriver";
}



void FilesystemPoolDriver::setStackInstance(StackInstance* si) throw (DmException)
{
  this->si_ = si;
}



void FilesystemPoolDriver::setSecurityContext(const SecurityContext* ctx) throw (DmException)
{
  // String => const char*
  if (this->fqans_ != NULL) {
    for (int i = 0; i < this->nFqans_; ++i)
      delete [] this->fqans_[i];
    delete [] this->fqans_;
  }
  
  this->nFqans_ = ctx->groups.size();
  this->fqans_  = new char* [this->nFqans_];
  for (int i = 0; i < this->nFqans_; ++i) {
    this->fqans_[i] = new char [ctx->groups[i].name.length() + 1];
    strcpy(this->fqans_[i], ctx->groups[i].name.c_str());
  }

  // Call DPM API
  wrapCall(dpm_client_setAuthorizationId(boost::any_cast<uid_t>(ctx->user["uid"]),
                                         boost::any_cast<gid_t>(ctx->groups[0]["gid"]),
                                         "GSI",
                                         (char*)ctx->user.name.c_str()));

  if (!ctx->groups.empty())
    wrapCall(dpm_client_setVOMS_data((char*)ctx->groups[0].name.c_str(),
                                     (char**)this->fqans_,
                                     ctx->groups.size()));

  // Store
  this->secCtx_ = ctx;

  // Id mechanism
  if (this->tokenUseIp_)
    this->userId_ = this->secCtx_->credentials.remoteAddress;
  else
    this->userId_ = this->secCtx_->credentials.clientName;
}



PoolHandler* FilesystemPoolDriver::createPoolHandler(const std::string& poolName) throw (DmException)
{
  return new FilesystemPoolHandler(this, poolName);
}



void FilesystemPoolDriver::toBeCreated(const Pool& pool) throw (DmException)
{
  // Need to create the pool here, or DPM won't see it!
  std::vector<boost::any> groups = pool.getVector("groups");
  struct dpm_pool dpmPool;
  std::memset(&dpmPool, 0, sizeof(dpmPool));
  
  if (groups.size() > 0) {
    dpmPool.gids = new gid_t [groups.size()];
    unsigned i;
    for (i = 0; i < groups.size(); ++i)
      dpmPool.gids[i] = Extensible::anyToUnsigned(groups[i]);
    dpmPool.nbgids = i;
  }
  else {
    dpmPool.gids = new gid_t [1];
    dpmPool.gids[0] = 0;
    dpmPool.nbgids  = 0;
  }   
  
  strncpy(dpmPool.poolname, pool.name.c_str(), sizeof(dpmPool.poolname));
  dpmPool.defsize         = pool.getLong("defsize");
  dpmPool.gc_start_thresh = pool.getLong("gc_start_thresh");
  dpmPool.gc_stop_thresh  = pool.getLong("gc_stop_thresh");
  dpmPool.def_lifetime    = pool.getLong("def_lifetime");
  dpmPool.defpintime      = pool.getLong("defpintime");
  dpmPool.max_lifetime    = pool.getLong("max_lifetime");
  dpmPool.maxpintime      = pool.getLong("maxpintime");
  dpmPool.ret_policy      = pool.getString("ret_policy")[0];
  dpmPool.s_type          = pool.getString("s_type")[0];
  
  strncpy(dpmPool.fss_policy, pool.getString("fss_policy").c_str(),
          sizeof(dpmPool.fss_policy));
  strncpy(dpmPool.gc_policy, pool.getString("gc_policy").c_str(),
          sizeof(dpmPool.gc_policy));
  strncpy(dpmPool.mig_policy, pool.getString("mig_policy").c_str(),
          sizeof(dpmPool.mig_policy));
  strncpy(dpmPool.rs_policy, pool.getString("rs_policy").c_str(),
          sizeof(dpmPool.rs_policy));

  try {
    wrapCall(dpm_addpool(&dpmPool));
    delete [] dpmPool.gids;
  }
  catch (...) {
    delete [] dpmPool.gids;
    throw;
  }
  
  // Pool added, now need the filesystems
  std::vector<boost::any> filesystems = pool.getVector("filesystems");
  
  for (unsigned i = 0; i < filesystems.size(); ++i) {
    Extensible fs = boost::any_cast<Extensible>(filesystems[i]);
    
    wrapCall(dpm_addfs(dpmPool.poolname,
                      (char*)fs.getString("server").c_str(),
                      (char*)fs.getString("fs").c_str(),
                      fs.getLong("status"),
                      fs.getLong("weight")));
  }
}



void FilesystemPoolDriver::justCreated(const Pool& pool) throw (DmException)
{
  // Nothing here
}



void FilesystemPoolDriver::update(const Pool& pool) throw (DmException)
{
  std::vector<boost::any> groups = pool.getVector("groups");
  struct dpm_pool dpmPool;
  std::memset(&dpmPool, 0, sizeof(dpmPool));
  
  if (groups.size() > 0) {
    dpmPool.gids = new gid_t [groups.size()];
    unsigned i;
    for (i = 0; i < groups.size(); ++i)
      dpmPool.gids[i] = Extensible::anyToUnsigned(groups[i]);
    dpmPool.nbgids = i;
  }
  else {
    dpmPool.gids = new gid_t [1];
    dpmPool.gids[0] = 0;
    dpmPool.nbgids  = 0;
  }
  
  strncpy(dpmPool.poolname, pool.name.c_str(), sizeof(dpmPool.poolname));
  dpmPool.defsize         = pool.getLong("defsize");
  dpmPool.gc_start_thresh = pool.getLong("gc_start_thresh");
  dpmPool.gc_stop_thresh  = pool.getLong("gc_stop_thresh");
  dpmPool.def_lifetime    = pool.getLong("def_lifetime");
  dpmPool.defpintime      = pool.getLong("defpintime");
  dpmPool.max_lifetime    = pool.getLong("max_lifetime");
  dpmPool.maxpintime      = pool.getLong("maxpintime");
  dpmPool.ret_policy      = pool.getString("ret_policy")[0];
  dpmPool.s_type          = pool.getString("s_type")[0];
  
  strncpy(dpmPool.fss_policy, pool.getString("fss_policy").c_str(),
          sizeof(dpmPool.fss_policy));
  strncpy(dpmPool.gc_policy, pool.getString("gc_policy").c_str(),
          sizeof(dpmPool.gc_policy));
  strncpy(dpmPool.mig_policy, pool.getString("mig_policy").c_str(),
          sizeof(dpmPool.mig_policy));
  strncpy(dpmPool.rs_policy, pool.getString("rs_policy").c_str(),
          sizeof(dpmPool.rs_policy));

  try {
    wrapCall(dpm_modifypool(&dpmPool));
    delete [] dpmPool.gids;
  }
  catch (...) {
    delete [] dpmPool.gids;
    throw;
  }
  
  // Pool updated, now need the filesystems
  std::vector<boost::any> filesystems = pool.getVector("filesystems");
  
  for (unsigned i = 0; i < filesystems.size(); ++i) {
    Extensible fs = boost::any_cast<Extensible>(filesystems[i]);
    
    wrapCall(dpm_modifyfs((char*)fs.getString("server").c_str(),
                          (char*)fs.getString("fs").c_str(),
                          fs.getLong("status"),
                          fs.getLong("weight")));
  }
}



void FilesystemPoolDriver::toBeDeleted(const Pool& pool) throw (DmException)
{
  // Delete filesystems
  int nFs;
  struct dpm_fs* dpmFs = NULL;
  
  if (dpm_getpoolfs((char*)pool.name.c_str(), &nFs, &dpmFs) == 0) {
    for (int i = 0; i < nFs; ++i)
      wrapCall(dpm_rmfs(dpmFs[i].server, dpmFs[i].fs));
    free(dpmFs);
  }
  else if (serrno != EINVAL) {
    ThrowExceptionFromSerrno(serrno);
  }
  
  // Delete it here
  try {
    wrapCall(dpm_rmpool((char*)pool.name.c_str()));
  }
  catch (DmException& e) {
    if (e.code() != DM_NO_SUCH_FILE) throw;
    throw DmException(DM_NO_SUCH_POOL, "Pool %s not found", pool.name.c_str());
  }
}



FilesystemPoolHandler::FilesystemPoolHandler(FilesystemPoolDriver* driver, const std::string& poolName):
  driver_(driver), poolName_(poolName)
{
  // Nothing
}



FilesystemPoolHandler::~FilesystemPoolHandler()
{
  // Nothing
}



std::string FilesystemPoolHandler::getPoolType(void) throw (DmException)
{
  return "filesystem";
}



std::string FilesystemPoolHandler::getPoolName(void) throw (DmException)
{
  return this->poolName_;
}



uint64_t FilesystemPoolHandler::getTotalSpace(void) throw (DmException)
{
  this->update();
  return this->total_;
}



uint64_t FilesystemPoolHandler::getFreeSpace(void) throw (DmException)
{
  this->update();
  return this->free_;
}



bool FilesystemPoolHandler::poolIsAvailable(bool write = true) throw (DmException)
{
  std::vector<dpm_fs> fs = this->getFilesystems(this->poolName_);
  
  for (unsigned i = 0; i < fs.size(); ++i) {
    if ((write && fs[i].status == 0) || (!write && fs[i].status != FS_DISABLED))
      return true;
  }
  
  return false;
}



bool FilesystemPoolHandler::replicaIsAvailable(const Replica& replica) throw (DmException)
{
  // No API call for getting one specific FS
  std::string pool = boost::any_cast<std::string>(replica["pool"]);
  std::vector<dpm_fs> fs = this->getFilesystems(pool);
  
  for (unsigned i = 0; i < fs.size(); ++i) {
    std::string filesystem = boost::any_cast<std::string>(replica["filesystem"]);
    if (filesystem == fs[i].fs && replica.server == fs[i].server) {
      return (fs[i].status != FS_DISABLED);
    }
  }

  return false;
}



void FilesystemPoolHandler::update() throw (DmException)
{
  int npools;
  struct dpm_pool *pool_array;

  if (dpm_getpools(&npools, &pool_array) != 0)
    ThrowExceptionFromSerrno(serrno);

  bool found = false;
  for (int i = 0; i < npools && !found; ++i) {
    if (this->poolName_ == pool_array[i].poolname) {
      found = true;
      
      this->total_ = pool_array[i].capacity;
      if (pool_array[i].free >= 0)
        this->free_  = pool_array[i].free;
      else
        this->free_ = 0;
    }
  }
  
  // Free
  for (int i = 0; i < npools; ++i)
    free(pool_array[i].gids);
  free(pool_array);

  // Failed?
  if (!found)
    throw DmException(DM_NO_SUCH_POOL, "Pool %s not found", this->poolName_.c_str());
}



Location FilesystemPoolHandler::whereToRead(const Replica& replica) throw (DmException)
{

  Url rloc(replica.rfn);
  
  Chunk single;
  
  single.host   = rloc.domain;
  single.path   = rloc.path;
  single.offset = 0;
  single.size   = this->driver_->si_->getINode()->extendedStat(replica.fileid).stat.st_size;
  
  single["token"] = dmlite::generateToken(this->driver_->userId_, rloc.path,
                                          this->driver_->tokenPasswd_,
                                          this->driver_->tokenLife_);
  
  return Location(1, single);
}



void FilesystemPoolHandler::removeReplica(const Replica& replica) throw (DmException)
{
  if (dpm_delreplica((char*)replica.rfn.c_str()) != 0)
    ThrowExceptionFromSerrno(serrno);
}



Location FilesystemPoolHandler::whereToWrite(const std::string& sfn) throw (DmException)
{
  struct dpm_putfilereq     reqfile;
  struct dpm_putfilestatus *statuses = NULL;
  int                       nReplies, wait;
  char                      token[CA_MAXDPMTOKENLEN + 1];
  std::string               spaceToken;

  reqfile.to_surl        = (char*)sfn.c_str();
  reqfile.f_type         = 'P';
  reqfile.lifetime       = 0;
  reqfile.requested_size = 0;
  reqfile.f_lifetime     = 0;
  reqfile.ret_policy     = '\0';
  reqfile.ac_latency     = '\0';
  reqfile.s_token[0]     = '\0';
  
  try {
    spaceToken = boost::any_cast<std::string>(this->driver_->si_->get("SpaceToken"));
  }
  catch (...) {
  }

  if (!spaceToken.empty()) {
    char **space_ids;

    RETRY(dpm_getspacetoken(spaceToken.c_str(), &nReplies, &space_ids),
          this->driver_->retryLimit_);
    
    strncpy(reqfile.s_token, space_ids[0], sizeof(reqfile.s_token));

    for(int i = 0; i < nReplies; ++i)
      free(space_ids[i]);
    free(space_ids);
  }

  try {
    // 4 on overwrite allows to add additional replicas
    // This way, dmlite can take care of handling the catalog creation in advance
    // No, this is not documented, done as seen on dpm_replicate.c
    char  rfio[]  = { "rfio" };
    char *rfiov[] = { rfio };
    dpm_put(1, &reqfile, 1, rfiov, (char *)"dmlite::adapter::put", 1,
            0, token, &nReplies, &statuses);;

    wait = statuses[0].status == DPM_QUEUED  ||
           statuses[0].status == DPM_RUNNING ||
           statuses[0].status == DPM_ACTIVE;

    while (wait) {
      if (dpm_getstatus_putreq(token, 1, &reqfile.to_surl, &nReplies, &statuses) < 0)
        ThrowExceptionFromSerrno(serrno);
      if (!nReplies)
        throw DmException(DM_PUT_ERROR, "Didn't get a destination from DPM");

      switch(statuses[0].status) {
        case DPM_FAILED: case DPM_ABORTED: case DPM_EXPIRED:
          throw DmException(DM_QUERY_FAILED, "The DPM put request failed");
          break;
        case DPM_READY: case DPM_SUCCESS:
          wait = 0;
          break;
      }
    }
    
    
    Url url(statuses[0].turl);
    dpm_free_pfilest(nReplies, statuses);
    
    // Return the location with the token
    url.path = dmlite::Url::normalizePath(url.path);
    
    Chunk single;
    
    single.host   = url.domain;
    single.path   = url.path;
    single.offset = 0;
    single.size   = 0;
    
    single["sfn"]      = sfn;
    single["dpmtoken"] = std::string(token);
    single["token"]    = dmlite::generateToken(this->driver_->userId_,
                                               url.path,
                                               this->driver_->tokenPasswd_,
                                               this->driver_->tokenLife_, true);
    return Location(1, single);
  }
  catch (...) {
    // On exceptions, free first!
    if (statuses != NULL)
      dpm_free_pfilest(nReplies, statuses);
    throw;
  } 
}



void FilesystemPoolHandler::doneWriting(const Replica& replica,
                                        const Extensible& extras) throw (DmException)
{
  struct dpm_filestatus     *statuses;
  int                        nReplies;
  std::string                sfn;
  std::string                token;
  
  // Need the sfn
  if (!extras.hasField("sfn"))
    throw DmException(DM_INVALID_VALUE, "sfn not present");
  sfn = extras.getString("sfn");
  
  // Need dpm token
  if (!extras.hasField("dpmtoken"))
    throw DmException(DM_INVALID_VALUE, "dpmtoken not present");
  token = extras.getString("dpmtoken");

  // Put done
  char *sfnArray[] = { (char*)sfn.c_str() };
  if (dpm_putdone((char*)token.c_str(), 1, sfnArray, &nReplies, &statuses) < 0)
    ThrowExceptionFromSerrno(serrno);

  dpm_free_filest(nReplies, statuses);
}



std::vector<dpm_fs> FilesystemPoolHandler::getFilesystems(const std::string& poolname) throw (DmException)
{
  std::vector<dpm_fs> fsV;
  int nfs;
  struct dpm_fs* fs_array;
    
  if (dpm_getpoolfs((char*)poolname.c_str(),  &nfs, &fs_array) != 0)
    ThrowExceptionFromSerrno(serrno);

  if (nfs == 0)
    throw DmException(DM_NO_SUCH_FS, "There are no filesystems inside the pool " + poolname);

  for (int i = 0; i < nfs; ++i) {
    fsV.push_back(fs_array[i]);
  }
  free(fs_array);
  
  return fsV;
}
