/// @file   IO.cpp
/// @brief  Filesystem IO.
/// @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
#include <dmlite/cpp/dmlite.h>
#include <dpm_api.h>
#include <errno.h>
#include <serrno.h>
#include "Adapter.h"
#include "DpmAdapter.h"
#include "IO.h"

using namespace dmlite;



StdIOFactory::StdIOFactory(): passwd_("default"), useIp_(true)
{
  setenv("CSEC_MECH", "ID", 1);
  dpm_client_resetAuthorizationId();
}



StdIOFactory::~StdIOFactory()
{
  // Nothing
}



void StdIOFactory::configure(const std::string& key, const std::string& value) throw (DmException)
{
  if (key == "TokenPassword") {
    this->passwd_ = value;
  }
  else if (key == "TokenId") {
    if (strcasecmp(value.c_str(), "ip") == 0)
      this->useIp_ = true;
    else
      this->useIp_ = false;
  }
  else if (key == "DpmHost" || key == "Host") {
    setenv("DPM_HOST", value.c_str(), 1);
    setenv("DPNS_HOST", value.c_str(), 1);
  }
  else
    throw DmException(DM_UNKNOWN_OPTION, key + " not known");
}



IODriver* StdIOFactory::createIODriver(PluginManager* pm) throw (DmException)
{
  return new StdIODriver(this->passwd_, this->useIp_);
}



StdIODriver::StdIODriver(std::string passwd, bool useIp):
  si_(0), secCtx_(0), passwd_(passwd), useIp_(useIp)
{
  // Nothing
}



StdIODriver::~StdIODriver()
{
  // Nothing
}



std::string StdIODriver::getImplId() const throw ()
{
  return "StdIODriver";
}



void StdIODriver::setStackInstance(StackInstance* si) throw (DmException)
{
  this->si_ = si;
}



void StdIODriver::setSecurityContext(const SecurityContext* ctx) throw (DmException)
{
  this->secCtx_ = ctx;
}



IOHandler* StdIODriver::createIOHandler(const std::string& pfn,
                                        int openmode,
                                        const Extensible& extras) throw (DmException)
{  
  if (!extras.hasField("token"))
    throw DmException(DM_FORBIDDEN, "Missing token");
  
  std::string userId;
  if (this->useIp_)
    userId = this->secCtx_->credentials.remoteAddress;
  else
    userId = this->secCtx_->credentials.clientName;
  
  if (dmlite::validateToken(extras.getString("token"),
                            userId,
                            pfn, this->passwd_,
                            openmode != O_RDONLY) != kTokenOK)
    throw DmException(DM_FORBIDDEN, "Token does not validate (using %s)",
                      this->useIp_?"IP":"DN");
  
  // Create
  return new StdIOHandler(pfn, openmode);
}



void StdIODriver::doneWriting(const std::string& pfn,
                              const Extensible& params) throw (DmException)
{
  struct dpm_filestatus     *statuses;
  int                        nReplies;
  std::string                sfn;
  
  // Need the sfn
  if (!params.hasField("sfn"))
    throw DmException(DM_INVALID_VALUE, "sfn not specified");
  sfn = params.getString("sfn");

  // Need the dpm token
  std::string token;
  if (!params.hasField("dpmtoken"))
    throw DmException(DM_INVALID_VALUE, "dpmtoken not specified"); 
  token = params.getString("dpmtoken");

  const char* sfnPtr[] = { sfn.c_str() };
  RETRY(dpm_putdone((char*)token.c_str(), 1, (char**)&sfnPtr, &nReplies, &statuses),
        3);

  dpm_free_filest(nReplies, statuses);
}



StdIOHandler::StdIOHandler(const std::string& path,
                           int openmode) throw (DmException):
  eof_(false), pos_(0)
{ 
  this->fd_ = ::open(path.c_str(), openmode, 0644);
  if (this->fd_ == -1)
    throw DmException(DM_NO_SUCH_FILE, "Could not open %s (%d)", path.c_str(), errno);
}



StdIOHandler::~StdIOHandler()
{
  if (this->fd_ != -1)
    ::close(this->fd_);
}



void StdIOHandler::close() throw (DmException)
{
  ::close(this->fd_);
  this->fd_ = -1;
}



size_t StdIOHandler::read(char* buffer, size_t count) throw (DmException)
{
  size_t nbytes = ::read(this->fd_, buffer, count);
  
  eof_ = (nbytes < count);
  
  return nbytes;
}



size_t StdIOHandler::write(const char* buffer, size_t count) throw (DmException)
{
  return ::write(this->fd_, buffer, count);
}



void StdIOHandler::seek(off_t offset, Whence whence) throw (DmException)
{
  if ((pos_ = ::lseek(this->fd_, offset, whence)) == ((off_t) - 1))
    throw DmException(DM_INTERNAL_ERROR, "Could not seek (%d)", errno);
}



long StdIOHandler::tell(void) throw (DmException)
{
  return pos_;
}



void StdIOHandler::flush(void) throw (DmException)
{
  // Nothing
}



bool StdIOHandler::eof(void) throw (DmException)
{
  return eof_;
}
