/// @file   NsAdapter.h
/// @brief  Adapter Plugin: Cns wrapper implementation.
/// @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
#ifndef NS_ADAPTER_H
#define	NS_ADAPTER_H

#include <dirent.h>
#include <dpns_api.h>
#include <dmlite/cpp/catalog.h>


namespace dmlite {
  
  const int kAclEntriesMax = 300;
  const int kAclSize       = 13;
  const int kCommentMax    = 255;

  /// @brief The private structure used by NsDummy to handle
  ///        openDir/readDir/closeDir
  struct PrivateDir: public Directory {
    virtual ~PrivateDir() {};
    dpns_DIR     *dpnsDir; ///< Used for calls to the dpns API.
    ExtendedStat  stat;    ///< Where the data is actually stored.
  };

  /// Catalog implemented as a wrapper around Cns API
  class NsAdapterCatalog: public Catalog, public Authn
  {
   public:
    /// Constructor
    /// @param retryLimit Limit of retrials.
    NsAdapterCatalog(unsigned retryLimit) throw (DmException);

    /// Destructor
    ~NsAdapterCatalog();

    // Overloading
    std::string getImplId(void) const throw ();

    void setStackInstance(StackInstance* si) throw (DmException);

    SecurityContext* createSecurityContext(const SecurityCredentials&) throw (DmException);
    void setSecurityContext(const SecurityContext*) throw (DmException);

    void        changeDir     (const std::string&) throw (DmException);
    std::string getWorkingDir (void)               throw (DmException);

    ExtendedStat extendedStat(const std::string&, bool) throw (DmException);

    SymLink readLink(ino_t) throw (DmException);

    void addReplica   (const Replica&) throw (DmException);
    void deleteReplica(const Replica&) throw (DmException);
    std::vector<Replica> getReplicas(const std::string&) throw (DmException);

    void        symlink (const std::string&, const std::string&) throw (DmException);
    std::string readLink(const std::string& path) throw (DmException);
    
    void unlink(const std::string&)                     throw (DmException);

    void create(const std::string&, mode_t) throw (DmException);

    mode_t umask   (mode_t)                           throw ();
    void   setMode (const std::string&, mode_t)       throw (DmException);
    void   setOwner(const std::string&, uid_t, gid_t, bool) throw (DmException);

    void setSize    (const std::string&, size_t) throw (DmException);
    void setChecksum(const std::string&,
                    const std::string&, const std::string&) throw (DmException);

    void setAcl(const std::string&, const Acl&) throw (DmException);

    void utime(const std::string&, const struct utimbuf*) throw (DmException);

    std::string getComment(const std::string&) throw (DmException);
    void        setComment(const std::string&,
                          const std::string&) throw (DmException);

    void setGuid(const std::string&, const std::string&) throw (DmException);

    void updateExtendedAttributes(const std::string&,
                                  const Extensible&) throw (DmException);
    
    GroupInfo getGroup   (gid_t)              throw (DmException);
    GroupInfo getGroup   (const std::string&) throw (DmException);
    GroupInfo getGroup   (const std::string& key,
                          const boost::any& value) throw (DmException);
    GroupInfo newGroup   (const std::string& gname) throw (DmException);
    void      updateGroup(const GroupInfo& group) throw (DmException);
    void      deleteGroup(const std::string& groupName) throw (DmException);
    
    UserInfo  getUser   (const std::string&) throw (DmException);
    UserInfo  getUser   (const std::string& key,
                         const boost::any& value) throw (DmException);
    UserInfo  newUser   (const std::string& uname) throw (DmException);
    void      updateUser(const UserInfo& user) throw (DmException);
    void      deleteUser(const std::string& userName) throw (DmException);
    
    std::vector<GroupInfo> getGroups(void) throw (DmException);
    std::vector<UserInfo>  getUsers (void) throw (DmException);

    Directory* openDir (const std::string&) throw (DmException);
    void       closeDir(Directory*)         throw (DmException);

    struct dirent* readDir (Directory*) throw (DmException);
    ExtendedStat*  readDirx(Directory*) throw (DmException);

    void makeDir(const std::string&, mode_t) throw (DmException);

    void rename     (const std::string&, const std::string&) throw (DmException);
    void removeDir  (const std::string&)                     throw (DmException);

    Replica getReplica   (const std::string& rfn)        throw (DmException);
    void    updateReplica(const Replica& replica) throw (DmException);
       
    void getIdMap(const std::string& userName,
                  const std::vector<std::string>& groupNames,
                  UserInfo* user,
                  std::vector<GroupInfo>* groups) throw (DmException);

   protected:
    StackInstance* si_;

    unsigned    retryLimit_;
    std::string cwdPath_;

    // Need to keep this in memory, as dpns/dpm API do not make
    // copy of them (sigh)
    // setAuthorizationId does, though (don't ask me why)
    char   **fqans_;
    size_t   nFqans_;
  };

};

#endif	// NS_ADAPTER_H
